/*
 * Possibly one of the most evil programs conceivable to seek out
 * and copy game data files to the current directory.
 *
 * Written in the afternoon and evening of the 7th August 2004
 * by Jonathon Fowler (jonof@edgenetwork.org)
 * http://jonof.edgenetwork.org
 *
 * This is released under the GNU General Public License. See GNU.TXT
 *
 * The icon I'm using with this was derived from one found in the Gnome
 * generic icon theme. http://www.gnome.org
 *
 * I wrote this to compile with MinGW32. I don't know if it'll work on
 * VisualC or OpenWatcom, and I'm not really worried at the moment about
 * that. I also tried to use Windows' TCHAR nonsense which should just
 * equate to 'char' but in theory I guess defining UNICODE and _UNICODE
 * before the windows.h include would make this Unicode... Let's not bet
 * on that though. ANSI works just fine for now.
 *
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <shlobj.h>
#include <commctrl.h>

#include <string.h>
#include <stdio.h>
#include <tchar.h>

#include "dialogs.h"
#include "defs.h"	// declares the versions of files we look for depending on a macro

LPITEMIDLIST lpuserdir = NULL;
TCHAR userlocname[MAX_PATH];
TCHAR currentdir[MAX_PATH];

#define SearchCD 1
#define SearchHD 2
#define SearchLoc 4
int whattosearch = 0;
int searchcancelled = 0;

HWND searchdlg = NULL;

int matches = 0;
#define MAXMATCHES 128
struct _finds {
	int type;
	TCHAR location[MAX_PATH];
} finds[MAXMATCHES];

TCHAR findroot[MAX_PATH*2];
int statusctl = 0;


BOOL DoCopy(TCHAR * srcpath, TCHAR * dstpath, HWND hwnd)
{
	TCHAR dest[MAX_PATH], src[MAX_PATH];

	_tcscpy(dest,dstpath);						
	_tcscat(dest,"\\"GRP);

	_tcscpy(src,srcpath);
	_tcscat(src,"\\"GRP);
					
	if (!CopyFile(src,dest,FALSE)) {
		MessageBox(hwnd, "Failed to copy " GRP ". The file was not installed.", "Copy failed", MB_OK|MB_ICONSTOP);
		return FALSE;
	} else {
		MessageBox(hwnd,GRP " was successfully copied.", "Copy succeeded", MB_OK|MB_ICONINFORMATION);
		return TRUE;
	}
}

void About(HWND hwnd)
{
	MessageBox(hwnd,"Game Data Installer\n"
			"This version built for "GAME" on "__DATE__" "__TIME__"\n"
			"(c)2004 Jonathon Fowler\n"
			"http://www.jonof.id.au/",
			"About",MB_OK);
}

INT_PTR CALLBACK DlgProc_Main(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
INT_PTR CALLBACK DlgProc_SearchRun(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
INT_PTR CALLBACK DlgProc_SearchFail(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
INT_PTR CALLBACK DlgProc_SelectGrp(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

INT_PTR CALLBACK DlgProc_Main(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
		case WM_INITDIALOG:
			EnableWindow(GetDlgItem(hwnd,MAIN_SEARCHLOCSTR),0);
			EnableWindow(GetDlgItem(hwnd,MAIN_SEARCHLOCPICK),0);
			break;
			
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case MAIN_ABOUT:
					About(hwnd);
					return TRUE;
				case MAIN_SEARCHLOC:
					{
						int cl;
						cl = IsDlgButtonChecked(hwnd,MAIN_SEARCHLOC) == BST_UNCHECKED ? FALSE : TRUE;
						EnableWindow(GetDlgItem(hwnd,MAIN_SEARCHLOCSTR),cl);
						EnableWindow(GetDlgItem(hwnd,MAIN_SEARCHLOCPICK),cl);
					}
					return TRUE;
				case MAIN_SEARCHLOCPICK:
					{
						LPITEMIDLIST lpiil;
						BROWSEINFO lpbi;
						TCHAR displayname[MAX_PATH];

						lpbi.hwndOwner = hwnd;
						lpbi.pidlRoot = lpuserdir;
						lpbi.pszDisplayName = displayname;
						lpbi.lpszTitle = "Choose a folder";
						lpbi.ulFlags = BIF_RETURNONLYFSDIRS;
						lpbi.lpfn = NULL;
						lpbi.lParam = 0;

						lpiil = SHBrowseForFolder(&lpbi);
						if (lpiil) {
							if (SHGetPathFromIDList(lpiil,displayname)) {
								userlocname[0] = 0;
								lpuserdir = NULL;
							} else lpuserdir = lpiil;
							SetDlgItemText(hwnd,MAIN_SEARCHLOCSTR,displayname);
							_tcscpy(userlocname,displayname);
						}
					}
					return TRUE;
				case IDOK:
					whattosearch = 0;
					whattosearch |= IsDlgButtonChecked(hwnd,MAIN_SEARCHCD) == BST_UNCHECKED ? 0 : SearchCD;
					whattosearch |= IsDlgButtonChecked(hwnd,MAIN_SEARCHHD) == BST_UNCHECKED ? 0 : SearchHD;
					whattosearch |= IsDlgButtonChecked(hwnd,MAIN_SEARCHLOC) == BST_UNCHECKED ? 0 : SearchLoc;
					if (!whattosearch) {
						MessageBox(hwnd,"You have not selected any search options. "
								"Please choose at least one to continue.",
								"No search selected",MB_OK|MB_ICONINFORMATION);
						return TRUE;
					}
				case IDCANCEL:
					EndDialog(hwnd,LOWORD(wParam));
					return TRUE;
			}
			break;
			
	}
	return FALSE;
}

INT_PTR CALLBACK DlgProc_SearchRun(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
		case WM_INITDIALOG:
			searchcancelled=0;
			SetDlgItemText(hwnd,SEARCH_CD,"");
			SetDlgItemText(hwnd,SEARCH_HD,"");
			SetDlgItemText(hwnd,SEARCH_LOC,"");
			return FALSE;
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDCANCEL:
					DestroyWindow(hwnd);
					searchdlg = NULL;
					searchcancelled = 1;
					return TRUE;
			}
			break;
	}
	return FALSE;
}

INT_PTR CALLBACK DlgProc_SearchFail(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	if (uMsg == WM_COMMAND && LOWORD(wParam) == IDOK) {
		EndDialog(hwnd,IDOK);
		return TRUE;
	}
	return FALSE;
}

INT_PTR CALLBACK DlgProc_SelectGrp(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
		case WM_INITDIALOG:
			{
				int i;
				TCHAR str[MAX_PATH*2];
				LRESULT listidx;
				
				ShowWindow(GetDlgItem(hwnd,SELECTGRP_MSG),SW_HIDE);
				for (i=0;i<matches;i++) {
					_sntprintf(str,MAX_PATH*2,"\"%s\" in %s",types[ finds[i].type ].name, finds[i].location);
					listidx = ListBox_AddString(GetDlgItem(hwnd,SELECTGRP_LIST),str);
					ListBox_SetItemData(GetDlgItem(hwnd,SELECTGRP_LIST),listidx,i);
				}
			}
			return TRUE;
		case WM_COMMAND:
			switch (LOWORD(wParam)) {
				case IDOK: {
					int item,cur;

					item = ListBox_GetCurSel(GetDlgItem(hwnd,SELECTGRP_LIST));
					if (item == LB_ERR) {
						MessageBox(hwnd,"Please select one of the items in the list before proceeding.","No item selected",MB_OK|MB_ICONINFORMATION);
						return TRUE;
					}

					cur = ListBox_GetItemData(GetDlgItem(hwnd,SELECTGRP_LIST),item);
					if (cur < 0 || cur >= matches) {
						MessageBox(hwnd,"The selected item wasn't found!?","Error",MB_OK|MB_ICONSTOP);
						return TRUE;
					}

					SetCursor(LoadCursor(NULL,IDC_WAIT));
					ShowWindow(GetDlgItem(hwnd,SELECTGRP_MSG),SW_SHOW);
					EnableWindow(GetDlgItem(hwnd,IDOK),FALSE);
					EnableWindow(GetDlgItem(hwnd,IDCANCEL),FALSE);
					EnableWindow(GetDlgItem(hwnd,SELECTGRP_LIST),FALSE);
					UpdateWindow(hwnd);
					
					DoCopy(finds[cur].location, currentdir, hwnd);
					
					SetCursor(LoadCursor(NULL,IDC_ARROW));
				}	// fall through
				case IDCANCEL:
					EndDialog(hwnd,LOWORD(wParam));
					return TRUE;
			}
			break;
	}
	return FALSE;
}


int DoSearch(void)
{
	int rv = 0,i,incurdir=0;
	MSG msg;

	TCHAR *end;
	HANDLE findh;
	WIN32_FIND_DATA found;

	end = &findroot[_tcslen(findroot)];
	findh = FindFirstFile(findroot,&found);
	if (findh == INVALID_HANDLE_VALUE) return 0;

	*(end-1) = 0;
	SetDlgItemText(searchdlg,statusctl,findroot);

	if (!_tcsicmp(findroot,currentdir)) incurdir = 1;

	do {
		if (matches == MAXMATCHES) break;

		if (PeekMessage(&msg,NULL,0,0,PM_REMOVE)) {
			if (!IsDialogMessage(searchdlg, &msg)) {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
		if (searchcancelled) { rv = -1; break; }
		
		if (found.dwFileAttributes & (FILE_ATTRIBUTE_SYSTEM|FILE_ATTRIBUTE_TEMPORARY|FILE_ATTRIBUTE_OFFLINE))
			continue;
		if (found.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
			if (!_tcscmp(found.cFileName,".") || !_tcscmp(found.cFileName,".."))
				continue;
			_tcscpy(end-1,found.cFileName);
			_tcscat(findroot,"\\*");
			DoSearch();
			*(end-1) = 0;
		} else if (!incurdir) {
			if (!_tcsicmp(found.cFileName,GRP)) {
				for (i=0;i<ntypes;i++)
					if (found.nFileSizeLow == types[i].size) break;
				if (i<ntypes) {
					_tcscpy(finds[matches].location,findroot);
					finds[matches].type = i;
					matches++;
				}
			}
		}
	} while (FindNextFile(findh,&found));
	FindClose(findh);
	
	return rv;
}

BOOL HasGOGVersionInstalled(void)
{
#ifdef DUKE3D
	HKEY key;
	DWORD type, size;

	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, TEXT("SOFTWARE\\GOG.com\\GOGDUKE3D"),
				0, KEY_READ, &key) != ERROR_SUCCESS) {
		return FALSE;
	}

	size = sizeof(userlocname);
	if (RegQueryValueEx(key, TEXT("PATH"), NULL, &type, (PBYTE)userlocname, &size) != ERROR_SUCCESS) {
		RegCloseKey(key);
		return FALSE;
	}

	RegCloseKey(key);

	if (type == REG_SZ) {
		return TRUE;
	}
#endif

	return FALSE;
}

int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nCmdShow)
{
	TCHAR drives[4*32+1], *drivep;
	BOOL done = FALSE;

	if (!GetLogicalDriveStrings(sizeof(drives)/sizeof(TCHAR),drives)) {
		MessageBox(NULL,"Failed getting drive letters.","Error",MB_OK|MB_ICONSTOP);
		return -1;
	}

	memset((void*)finds,0,sizeof(finds));

	// evil and nasty but it ought to work
	GetCurrentDirectory(MAX_PATH,currentdir);
	_tcscat(currentdir,"\\");
	
	if (FAILED(CoInitialize(NULL))) {
		MessageBox(NULL,"Failed initialising COM","COM Error",MB_OK|MB_ICONSTOP);
		return -1;
	}

	InitCommonControls();

	if (HasGOGVersionInstalled()) {
		if (MessageBox(NULL, "A copy of " ORIGGAME " installed from GOG.com has been detected. "
				"Do you wish to use the game data from this copy with " GAME "?",
				GAME " Game Data Installer", MB_YESNO|MB_ICONQUESTION) == IDYES) {
			done = DoCopy(userlocname, currentdir, NULL);
		}

		userlocname[0] = 0;
	}
	
	if (!done && DialogBox(hInst,MAKEINTRESOURCE(DIALOG_MAIN),NULL,DlgProc_Main) == IDOK) {

		searchdlg = CreateDialog(hInst,MAKEINTRESOURCE(DIALOG_SEARCHRUN),NULL,DlgProc_SearchRun);
		if (searchdlg) {
			matches = 0;

			if (!(whattosearch & SearchCD)) {
				SetDlgItemText(searchdlg,SEARCH_CD,"Skipped");
			} else if (!searchcancelled) {
				for (drivep = drives; *drivep; drivep += _tcslen(drivep)+1) {
					if (GetDriveType(drivep) != DRIVE_CDROM) continue;
					_tcscpy(findroot,drivep); _tcscat(findroot,"*");
					statusctl = SEARCH_CD;
					if (DoSearch()) goto earlyexit;
				}
				SetDlgItemText(searchdlg,SEARCH_CD,"Done");
			}
			
			if (!(whattosearch & SearchHD)) {
				SetDlgItemText(searchdlg,SEARCH_HD,"Skipped");
			} else if (!searchcancelled) {
				for (drivep = drives; *drivep; drivep += _tcslen(drivep)+1) {
					if (GetDriveType(drivep) != DRIVE_FIXED) continue;
					_tcscpy(findroot,drivep); _tcscat(findroot,"*");
					statusctl = SEARCH_HD;
					if (DoSearch()) goto earlyexit;
				}
				SetDlgItemText(searchdlg,SEARCH_HD,"Done");
			}
			
			if (!(whattosearch & SearchLoc) || !userlocname[0]) {
				SetDlgItemText(searchdlg,SEARCH_LOC,"Skipped");
			} else if (!searchcancelled) {
				_tcscpy(findroot,userlocname);
				if (findroot[ _tcslen(findroot) - 1 ] != '\\')
					_tcscat(findroot,"\\");
				_tcscat(findroot,"*");
				statusctl = SEARCH_LOC;
				if (DoSearch()) goto earlyexit;
				SetDlgItemText(searchdlg,SEARCH_LOC,"Done");
			}

earlyexit:
			if (searchdlg) {
				DestroyWindow(searchdlg);
				searchdlg = NULL;
			}

			if (searchcancelled) {
				MessageBox(NULL,"The search has been cancelled. Exiting.",
						"Search aborted", MB_OK|MB_ICONEXCLAMATION);
			} else {
				if (matches) {
					DialogBox(hInst,MAKEINTRESOURCE(DIALOG_SELECTGRP),NULL,DlgProc_SelectGrp);
				} else {
					DialogBox(hInst,MAKEINTRESOURCE(DIALOG_SEARCHFAIL),NULL,DlgProc_SearchFail);
				}
			}
		}
		
	}

	// ???: should we be freeing the lpuserdir variable or what?

	CoUninitialize();

	return 0;
}

